@@ -21,6 +21,8 @@ Database models of the Dashboard application
 """
 
 import datetime
+import errno
+import gzip
 import hashlib
 import logging
 import os
@@ -28,10 +30,13 @@ import simplejson
 import traceback
 import contextlib
 
+from django.conf import settings
 from django.contrib.auth.models import User
 from django.contrib.contenttypes import generic
 from django.contrib.contenttypes.models import ContentType
 from django.core.exceptions import ValidationError
+from django.core.files import locks, File
+from django.core.files.storage import FileSystemStorage
 from django.core.urlresolvers import reverse
 from django.db import models
 from django.template import Template, Context
@@ -311,6 +316,62 @@ class BundleStream(RestrictedResource):
         return user, group, slug, is_public, is_anonymous
 
 
+class GzipFileSystemStorage(FileSystemStorage):
+
+    def _open(self, name, mode='rb'):
+        full_path = self.path(name)
+        gzip_file = gzip.GzipFile(full_path, mode)
+        gzip_file.name = full_path
+        return File(gzip_file)
+
+    # This is a copy-paste-hack of FileSystemStorage._save
+    def _save(self, name, content):
+        full_path = self.path(name)
+
+        directory = os.path.dirname(full_path)
+        if not os.path.exists(directory):
+            os.makedirs(directory)
+        elif not os.path.isdir(directory):
+            raise IOError("%s exists and is not a directory." % directory)
+
+        # There's a potential race condition between get_available_name and
+        # saving the file; it's possible that two threads might return the
+        # same name, at which point all sorts of fun happens. So we need to
+        # try to create the file, but if it already exists we have to go back
+        # to get_available_name() and try again.
+
+        while True:
+            try:
+                # This fun binary flag incantation makes os.open throw an
+                # OSError if the file already exists before we open it.
+                fd = os.open(full_path, os.O_WRONLY | os.O_CREAT | os.O_EXCL | getattr(os, 'O_BINARY', 0))
+                # This line, and the use of gz_file.write below, are the
+                # changes from the original version of this.
+                gz_file = gzip.GzipFile(fileobj=os.fdopen(fd, 'wb'))
+                try:
+                    locks.lock(fd, locks.LOCK_EX)
+                    for chunk in content.chunks():
+                        gz_file.write(chunk)
+                finally:
+                    locks.unlock(fd)
+                    gz_file.close()
+            except OSError, e:
+                if e.errno == errno.EEXIST:
+                    # Ooops, the file exists. We need a new file name.
+                    name = self.get_available_name(name)
+                    full_path = self.path(name)
+                else:
+                    raise
+            else:
+                # OK, the file save worked. Break out of the loop.
+                break
+
+        if settings.FILE_UPLOAD_PERMISSIONS is not None:
+            os.chmod(full_path, settings.FILE_UPLOAD_PERMISSIONS)
+
+        return name
+
+
 class Bundle(models.Model):
     """
     Model for "Dashboard Bundles"
@@ -337,11 +398,34 @@ class Bundle(models.Model):
                 " into the database"),
             editable = False)
 
-    content = models.FileField(
+    _raw_content = models.FileField(
             verbose_name = _(u"Content"),
             help_text = _(u"Document in Dashboard Bundle Format 1.0"),
             upload_to = 'bundles',
-            null = True)
+            null = True,
+            db_column = 'content')
+
+    _gz_content = models.FileField(
+            verbose_name = _(u"Compressed content"),
+            help_text = _(u"Compressed document in Dashboard Bundle Format 1.0"),
+            upload_to = 'compressed-bundles',
+            null = True,
+            db_column = 'gz_content',
+            storage = GzipFileSystemStorage())
+
+    def _get_content(self):
+        r = self._gz_content
+        if not r:
+            return self._raw_content
+        else:
+            return r
+
+    content = property(_get_content)
+
+    def compress(self):
+        c = self._raw_content
+        self._gz_content.save(c.name, c)
+        c.delete()
 
     content_sha1 = models.CharField(
             editable = False,
